@@ -12,15 +12,28 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from typing import Callable, Optional, Tuple
+from typing import Optional, Tuple
 
 import tensorflow as tf
 
 from ..base import MeanAndVariance
 from ..config import default_float, default_jitter
+from ..experimental.check_shapes import check_shape as cs
+from ..experimental.check_shapes import check_shapes
 from ..utilities.ops import leading_transpose
 
 
+@check_shapes(
+    "Kmn: [M, batch..., N]",
+    "Kmm: [M, M]",
+    "Knn: [batch..., N, N] if full_cov",
+    "Knn: [batch..., N] if not full_cov",
+    "f: [M, R]",
+    "q_sqrt: [M_R_or_R_M_M...]",
+    "return[0]: [batch..., N, R]",
+    "return[1]: [batch..., R, N, N] if full_cov",
+    "return[1]: [batch..., N, R] if not full_cov",
+)
 def base_conditional(
     Kmn: tf.Tensor,
     Kmm: tf.Tensor,
@@ -47,15 +60,9 @@ def base_conditional(
 
       q(g1) = ∫ q(g2) p(g1 | g2)
 
-    :param Kmn: [M, ..., N]
-    :param Kmm: [M, M]
-    :param Knn: [..., N, N]  or  N
-    :param f: [M, R]
-    :param full_cov: bool
     :param q_sqrt: If this is a Tensor, it must have shape [R, M, M] (lower
         triangular) or [M, R] (diagonal)
-    :param white: bool
-    :return: [N, R]  or [R, N, N]
+    :return: mean, variance
     """
     Lm = tf.linalg.cholesky(Kmm)
     return base_conditional_with_lm(
@@ -63,6 +70,17 @@ def base_conditional(
     )
 
 
+@check_shapes(
+    "Kmn: [M, batch..., N]",
+    "Lm: [M, M]",
+    "Knn: [batch..., N, N] if full_cov",
+    "Knn: [batch..., N] if not full_cov",
+    "f: [M, R]",
+    "q_sqrt: [M_R_or_R_M_M...]",
+    "return[0]: [batch..., N, R]",
+    "return[1]: [batch..., R, N, N] if full_cov",
+    "return[1]: [batch..., N, R] if not full_cov",
+)
 def base_conditional_with_lm(
     Kmn: tf.Tensor,
     Lm: tf.Tensor,
@@ -79,6 +97,9 @@ def base_conditional_with_lm(
 
     This allows `Lm` to be precomputed, which can improve performance.
     """
+    if q_sqrt is not None:
+        cs(q_sqrt, "[M, R]" if q_sqrt.shape.ndims == 2 else "[R, M, M]")
+
     # compute kernel stuff
     num_func = tf.shape(f)[-1]  # R
     N = tf.shape(Kmn)[-1]
@@ -97,24 +118,6 @@ def base_conditional_with_lm(
     )  # [N]
     Kmn = tf.transpose(Kmn, perm)  # [..., M, N]
 
-    shape_constraints = [
-        (Kmn, [..., "M", "N"]),
-        (Lm, ["M", "M"]),
-        (Knn, [..., "N", "N"] if full_cov else [..., "N"]),
-        (f, ["M", "R"]),
-    ]
-    if q_sqrt is not None:
-        shape_constraints.append(
-            (q_sqrt, (["M", "R"] if q_sqrt.shape.ndims == 2 else ["R", "M", "M"]))
-        )
-    tf.debugging.assert_shapes(
-        shape_constraints,
-        message="base_conditional() arguments "
-        "[Note that this check verifies the shape of an alternative "
-        "representation of Kmn. See the docs for the actual expected "
-        "shape.]",
-    )
-
     leading_dims = tf.shape(Kmn)[:-2]
 
     # Compute the projection matrix A
@@ -163,36 +166,24 @@ def base_conditional_with_lm(
     if not full_cov:
         fvar = tf.linalg.adjoint(fvar)  # [N, R]
 
-    shape_constraints = [
-        (Kmn, [..., "M", "N"]),  # tensor included again for N dimension
-        (f, [..., "M", "R"]),  # tensor included again for R dimension
-        (fmean, [..., "N", "R"]),
-        (fvar, [..., "R", "N", "N"] if full_cov else [..., "N", "R"]),
-    ]
-    tf.debugging.assert_shapes(shape_constraints, message="base_conditional() return values")
-
     return fmean, fvar
 
 
+@check_shapes(
+    "mean: [batch..., N, D]",
+    "cov: [batch..., N, D, D] if full_cov",
+    "cov: [batch..., N, D] if not full_cov",
+    "return: [batch..., N, D] if num_samples is None",
+    "return: [batch..., S, N, D] if num_samples is not None",
+)
 def sample_mvn(
     mean: tf.Tensor, cov: tf.Tensor, full_cov: bool, num_samples: Optional[int] = None
 ) -> tf.Tensor:
     """
     Returns a sample from a D-dimensional Multivariate Normal distribution.
 
-    :param mean: [..., N, D]
-    :param cov: [..., N, D] or [..., N, D, D]
-    :param full_cov: if `True` return a "full" covariance matrix, otherwise a "diag":
-        - "full": cov holds the full covariance matrix (without jitter)
-        - "diag": cov holds the diagonal elements of the covariance matrix
-    :return: sample from the MVN of shape [..., (S), N, D], S = num_samples
+    :return: sample from the MVN
     """
-    shape_constraints = [
-        (mean, [..., "N", "D"]),
-        (cov, [..., "N", "D", "D"] if full_cov else [..., "N", "D"]),
-    ]
-    tf.debugging.assert_shapes(shape_constraints, message="sample_mvn() arguments")
-
     mean_shape = tf.shape(mean)
     S = num_samples if num_samples is not None else 1
     D = mean_shape[-1]
@@ -215,31 +206,25 @@ def sample_mvn(
         samples = mean[..., None] + tf.linalg.matmul(chol, eps)  # [..., N, D, S]
         samples = leading_transpose(samples, [..., -1, -3, -2])  # [..., S, N, D]
 
-    shape_constraints = [
-        (mean, [..., "N", "D"]),
-        (samples, [..., "S", "N", "D"]),
-    ]
-    tf.debugging.assert_shapes(shape_constraints, message="sample_mvn() return values")
-
     if num_samples is None:
         return tf.squeeze(samples, axis=-3)  # [..., N, D]
     return samples  # [..., S, N, D]
 
 
+@check_shapes(
+    "fvar: [batch..., P, N, N] if full_cov",
+    "fvar: [batch..., N, P] if not full_cov",
+    "return: [batch..., N, P, N, P] if full_cov and full_output_cov",
+    "return: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+    "return: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+    "return: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+)
 def expand_independent_outputs(fvar: tf.Tensor, full_cov: bool, full_output_cov: bool) -> tf.Tensor:
     """
     Reshapes fvar to the correct shape, specified by `full_cov` and `full_output_cov`.
 
-    :param fvar: has shape [N, P] (full_cov = False) or [P, N, N] (full_cov = True).
-    :return:
-        1. full_cov: True and full_output_cov: True
-           fvar [N, P, N, P]
-        2. full_cov: True and full_output_cov: False
-           fvar [P, N, N]
-        3. full_cov: False and full_output_cov: True
-           fvar [N, P, P]
-        4. full_cov: False and full_output_cov: False
-           fvar [N, P]
+    :param fvar: Single-output covariance.
+    :return: Multi-output covariance.
     """
     if full_cov and full_output_cov:
         fvar = tf.linalg.diag(tf.transpose(fvar))  # [N, N, P, P]
@@ -254,6 +239,21 @@ def expand_independent_outputs(fvar: tf.Tensor, full_cov: bool, full_output_cov:
     return fvar
 
 
+@check_shapes(
+    "Kmn: [M, L, N, P]",
+    "Kmm: [L, M, M]",
+    "Knn: [N, P] if (not full_cov) and (not full_output_cov)",
+    "Knn: [P, N, N] if full_cov and (not full_output_cov)",
+    "Knn: [N, P, P] if (not full_cov) and full_output_cov",
+    "Knn: [N, P, N, P] if full_cov and full_output_cov",
+    "f: [M, L]",
+    "q_sqrt: [M_L_or_L_M_M...]",
+    "return[0]: [N, P]",
+    "return[1]: [N, P] if (not full_cov) and (not full_output_cov)",
+    "return[1]: [P, N, N] if full_cov and (not full_output_cov)",
+    "return[1]: [N, P, P] if (not full_cov) and full_output_cov",
+    "return[1]: [N, P, N, P] if full_cov and full_output_cov",
+)
 def independent_interdomain_conditional(
     Kmn: tf.Tensor,
     Kmm: tf.Tensor,
@@ -270,30 +270,13 @@ def independent_interdomain_conditional(
 
     Interdomain conditional calculation.
 
-    :param Kmn: [M, L, N, P]
-    :param Kmm: [L, M, M]
-    :param Knn: [N, P]  or  [N, P, P]  or  [P, N, N]  or  [N, P, N, P]
-    :param f: data matrix, [M, L]
-    :param q_sqrt: [L, M, M]  or  [M, L]
     :param full_cov: calculate covariance between inputs
     :param full_output_cov: calculate covariance between outputs
     :param white: use whitened representation
-    :return:
-        - mean: [N, P]
-        - variance: [N, P], [N, P, P], [P, N, N], [N, P, N, P]
+    :return: mean, variance
     """
     M, L, N, P = tf.unstack(tf.shape(Kmn), num=Kmn.shape.ndims, axis=0)
 
-    shape_constraints = [
-        (Kmn, ["M", "L", "N", "P"]),
-        (Kmm, ["L", "M", "M"]),
-        (f, ["M", "L"]),
-    ]
-    if q_sqrt is not None:
-        shape_constraints.append(
-            (q_sqrt, ["M", "L"] if q_sqrt.shape.ndims == 2 else ["L", "M", "M"])
-        )
-
     Lm = tf.linalg.cholesky(Kmm)  # [L, M, M]
 
     # Compute the projection matrix A
@@ -304,18 +287,14 @@ def independent_interdomain_conditional(
     # compute the covariance due to the conditioning
     if full_cov and full_output_cov:
         fvar = Knn - tf.tensordot(Ar, Ar, [[0, 1], [0, 1]])  # [N, P, N, P]
-        intended_cov_shape = ["N", "P", "N", "P"]
     elif full_cov and not full_output_cov:
         At = tf.reshape(tf.transpose(Ar), (P, N, M * L))  # [P, N, L]
         fvar = Knn - tf.linalg.matmul(At, At, transpose_b=True)  # [P, N, N]
-        intended_cov_shape = ["P", "N", "N"]
     elif not full_cov and full_output_cov:
         At = tf.reshape(tf.transpose(Ar, [2, 3, 1, 0]), (N, P, M * L))  # [N, P, L]
         fvar = Knn - tf.linalg.matmul(At, At, transpose_b=True)  # [N, P, P]
-        intended_cov_shape = ["N", "P", "P"]
     elif not full_cov and not full_output_cov:
         fvar = Knn - tf.reshape(tf.reduce_sum(tf.square(A), [0, 1]), (N, P))  # Knn: [N, P]
-        intended_cov_shape = ["N", "P"]
 
     # another backsubstitution in the unwhitened case
     if not white:
@@ -347,18 +326,24 @@ def independent_interdomain_conditional(
         elif not full_cov and not full_output_cov:
             fvar = fvar + tf.reshape(tf.reduce_sum(tf.square(LTA), (0, 1)), (N, P))
 
-    shape_constraints.extend(
-        [
-            (Knn, intended_cov_shape),
-            (fmean, ["N", "P"]),
-            (fvar, intended_cov_shape),
-        ]
-    )
-    tf.debugging.assert_shapes(shape_constraints, message="independent_interdomain_conditional()")
-
     return fmean, fvar
 
 
+@check_shapes(
+    "Kmn: [M, N, P]",
+    "Kmm: [M, M]",
+    "Knn: [N, P] if (not full_cov) and (not full_output_cov)",
+    "Knn: [P, N, N] if full_cov and (not full_output_cov)",
+    "Knn: [N, P, P] if (not full_cov) and full_output_cov",
+    "Knn: [N, P, N, P] if full_cov and full_output_cov",
+    "f: [M, 1]",
+    "q_sqrt: [_1_L_or_1_M_M...]",
+    "return[0]: [N, P]",
+    "return[1]: [N, P] if (not full_cov) and (not full_output_cov)",
+    "return[1]: [P, N, N] if full_cov and (not full_output_cov)",
+    "return[1]: [N, P, P] if (not full_cov) and full_output_cov",
+    "return[1]: [N, P, N, P] if full_cov and full_output_cov",
+)
 def fully_correlated_conditional(
     Kmn: tf.Tensor,
     Kmm: tf.Tensor,
@@ -374,17 +359,10 @@ def fully_correlated_conditional(
     This function handles conditioning of multi-output GPs in the case where the conditioning
     points are all fully correlated, in both the prior and posterior.
 
-    :param Kmn: [M, N, P]
-    :param Kmm: [M, M]
-    :param Knn: [N, P] or [N, P, N, P]
-    :param f: data matrix, [M, 1]
-    :param q_sqrt: [1, M, M] or [1, L]
     :param full_cov: calculate covariance between inputs
     :param full_output_cov: calculate covariance between outputs
     :param white: use whitened representation
-    :return:
-        - mean: [N, P]
-        - variance: [N, P], [N, P, P], [P, N, N], [N, P, N, P]
+    :return: mean, variance
     """
     mean, var = fully_correlated_conditional_repeat(
         Kmn,
@@ -399,6 +377,21 @@ def fully_correlated_conditional(
     return tf.squeeze(mean, axis=0), tf.squeeze(var, axis=0)
 
 
+@check_shapes(
+    "Kmn: [M, N, P]",
+    "Kmm: [M, M]",
+    "Knn: [N, P] if (not full_cov) and (not full_output_cov)",
+    "Knn: [P, N, N] if full_cov and (not full_output_cov)",
+    "Knn: [N, P, P] if (not full_cov) and full_output_cov",
+    "Knn: [N, P, N, P] if full_cov and full_output_cov",
+    "f: [M, R]",
+    "q_sqrt: [M_R_or_R_M_M...]",
+    "return[0]: [R, N, P]",
+    "return[1]: [R, N, P] if (not full_cov) and (not full_output_cov)",
+    "return[1]: [R, P, N, N] if full_cov and (not full_output_cov)",
+    "return[1]: [R, N, P, P] if (not full_cov) and full_output_cov",
+    "return[1]: [R, N, P, N, P] if full_cov and full_output_cov",
+)
 def fully_correlated_conditional_repeat(
     Kmn: tf.Tensor,
     Kmm: tf.Tensor,
@@ -415,31 +408,14 @@ def fully_correlated_conditional_repeat(
     points are all fully correlated, in both the prior and posterior.
     Note: This conditional can handle 'repetitions' R, given in `f` and `q_sqrt`.
 
-    :param Kmn: [M, N, P]
-    :param Kmm: [M, M]
-    :param Knn: [N, P] or [N, P, P] or [P, N, N] or [N, P, N, P]
-    :param f: data matrix, [M, R]
-    :param q_sqrt: [R, M, M] or [M, R]
     :param full_cov: calculate covariance between inputs
     :param full_output_cov: calculate covariance between outputs
     :param white: use whitened representation
-    :return:
-        - mean: [R, N, P]
-        - variance: [R, N, P], [R, N, P, P], [R, P, N, N], [R, N, P, N, P]
+    :return: mean, variance
     """
     R = tf.shape(f)[1]
     M, N, P = tf.unstack(tf.shape(Kmn), num=Kmn.shape.ndims, axis=0)
 
-    shape_constraints = [
-        (Kmn, ["M", "N", "P"]),
-        (Kmm, ["M", "M"]),
-        (f, ["M", "R"]),
-    ]
-    if q_sqrt is not None:
-        shape_constraints.append(
-            (q_sqrt, ["M", "R"] if q_sqrt.shape.ndims == 2 else ["R", "M", "M"])
-        )
-
     Lm = tf.linalg.cholesky(Kmm)
 
     # Compute the projection matrix A
@@ -452,22 +428,18 @@ def fully_correlated_conditional_repeat(
     if full_cov and full_output_cov:
         # fvar = Knn - tf.linalg.matmul(Ar, Ar, transpose_a=True)  # [P, P], then reshape?
         fvar = Knn - tf.tensordot(Ar, Ar, [[0], [0]])  # [N, P, N, P]
-        intended_cov_shape = ["N", "P", "N", "P"]
     elif full_cov and not full_output_cov:
         At = tf.transpose(Ar)  # [P, N, M]
         fvar = Knn - tf.linalg.matmul(At, At, transpose_b=True)  # [P, N, N]
-        intended_cov_shape = ["P", "N", "N"]
     elif not full_cov and full_output_cov:
         # This transpose is annoying
         At = tf.transpose(Ar, [1, 0, 2])  # [N, M, P]
         # fvar = Knn - tf.einsum('mnk,mnl->nkl', Ar, Ar)
         fvar = Knn - tf.linalg.matmul(At, At, transpose_a=True)  # [N, P, P]
-        intended_cov_shape = ["N", "P", "P"]
     elif not full_cov and not full_output_cov:
         # Knn: [N, P]
         # Can also do this with a matmul
         fvar = Knn - tf.reshape(tf.reduce_sum(tf.square(A), [0]), (N, P))
-        intended_cov_shape = ["N", "P"]
 
     # another backsubstitution in the unwhitened case
     if not white:
@@ -505,18 +477,13 @@ def fully_correlated_conditional_repeat(
         fvar_shape = tf.concat([[R], tf.shape(fvar)], axis=0)
         fvar = tf.broadcast_to(fvar[None], fvar_shape)
 
-    shape_constraints.extend(
-        [
-            (Knn, intended_cov_shape),
-            (fmean, ["R", "N", "P"]),
-            (fvar, ["R"] + intended_cov_shape),
-        ]
-    )
-    tf.debugging.assert_shapes(shape_constraints, message="fully_correlated_conditional_repeat()")
-
     return fmean, fvar
 
 
+@check_shapes(
+    "A: [left..., right...]",
+    "return: [right..., left...]",
+)
 def rollaxis_left(A: tf.Tensor, num_rolls: int) -> tf.Tensor:
     """Roll the tensor `A` backwards `num_rolls` times."""
     assert num_rolls > 0
@@ -525,6 +492,10 @@ def rollaxis_left(A: tf.Tensor, num_rolls: int) -> tf.Tensor:
     return tf.transpose(A, perm)
 
 
+@check_shapes(
+    "A: [left..., right...]",
+    "return: [right..., left...]",
+)
 def rollaxis_right(A: tf.Tensor, num_rolls: int) -> tf.Tensor:
     """Roll the tensor `A` forward `num_rolls` times."""
     assert num_rolls > 0
@@ -533,70 +504,65 @@ def rollaxis_right(A: tf.Tensor, num_rolls: int) -> tf.Tensor:
     return tf.transpose(A, perm)
 
 
+@check_shapes(
+    "W: [P, L]",
+    "g_mean: [batch..., N, L]",
+    "g_var: [batch..., N, L] if not full_cov",
+    "g_var: [L, batch..., N, N] if full_cov",
+    "return[0]: [batch..., N, P]",
+    "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+    "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+    "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+)
 def mix_latent_gp(
     W: tf.Tensor, g_mean: tf.Tensor, g_var: tf.Tensor, full_cov: bool, full_output_cov: bool
 ) -> MeanAndVariance:
     r"""Takes the mean and variance of an uncorrelated L-dimensional latent GP
     and returns the mean and the variance of the mixed GP, `f = W g`,
-    where both f and g are GPs, with W having a shape [P, L]
+    where both f and g are GPs.
 
-    :param W: [P, L]
-    :param g_mean: [..., N, L]
-    :param g_var: [..., N, L] (full_cov = False) or [L, ..., N, N] (full_cov = True)
-    :return: f_mean and f_var, shape depends on `full_cov` and `full_output_cov`
+    :return: f_mean and f_var
     """
-    shape_constraints = [
-        (W, ["P", "L"]),
-        (g_mean, [..., "N", "L"]),
-    ]
-    if not full_cov:
-        shape_constraints.append((g_var, [..., "N", "L"]))
-    else:
-        # NOTE(awav) cannot assert g_var shape here because of the inner "leading"
-        # dimensions, see https://github.com/GPflow/GPflow/issues/1296
-        pass
-
     f_mean = tf.tensordot(g_mean, W, [[-1], [-1]])  # [..., N, P]
 
     if full_cov and full_output_cov:  # g_var is [L, ..., N, N]
         # this branch is practically never taken
         g_var = rollaxis_left(g_var, 1)  # [..., N, N, L]
-        shape_constraints.append((g_var, [..., "N", "N", "L"]))
 
         g_var = tf.expand_dims(g_var, axis=-2)  # [..., N, N, 1, L]
         g_var_W = g_var * W  # [..., N, P, L]
         f_var = tf.tensordot(g_var_W, W, [[-1], [-1]])  # [..., N, N, P, P]
         f_var = leading_transpose(f_var, [..., -4, -2, -3, -1])  # [..., N, P, N, P]
-        intended_cov_shape = [..., "N", "P", "N", "P"]
 
     elif full_cov and not full_output_cov:  # g_var is [L, ..., N, N]
         # this branch is practically never taken
         f_var = tf.tensordot(g_var, W ** 2, [[0], [-1]])  # [..., N, N, P]
         f_var = leading_transpose(f_var, [..., -1, -3, -2])  # [..., P, N, N]
-        intended_cov_shape = [..., "P", "N", "N"]
 
     elif not full_cov and full_output_cov:  # g_var is [..., N, L]
         g_var = tf.expand_dims(g_var, axis=-2)  # [..., N, 1, L]
         g_var_W = g_var * W  # [..., N, P, L]
         f_var = tf.tensordot(g_var_W, W, [[-1], [-1]])  # [..., N, P, P]
-        intended_cov_shape = [..., "N", "P", "P"]
 
     elif not full_cov and not full_output_cov:  # g_var is [..., N, L]
         W_squared = W ** 2  # [P, L]
         f_var = tf.tensordot(g_var, W_squared, [[-1], [-1]])  # [..., N, P]
-        intended_cov_shape = [..., "N", "P"]
-
-    shape_constraints.extend(
-        [
-            (f_mean, [..., "N", "P"]),
-            (f_var, intended_cov_shape),
-        ]
-    )
-    tf.debugging.assert_shapes(shape_constraints, message="mix_latent_gp()")
 
     return f_mean, f_var
 
 
+@check_shapes(
+    "Kmns: [P, M, batch..., N]",
+    "Kmms: [P, M, M]",
+    "Knns: [P, batch..., N, N] if full_cov",
+    "Knns: [P, batch..., N] if not full_cov",
+    "f: [M, P]",
+    "q_sqrt: [M_R_or_R_M_M...]",
+    "return[0]: [batch..., N, R]",
+    "return[1]: [batch..., R, N, N] if full_cov",
+    "return[1]: [batch..., N, R] if not full_cov",
+)
 def separate_independent_conditional_implementation(
     Kmns: tf.Tensor,
     Kmms: tf.Tensor,
@@ -611,11 +577,6 @@ def separate_independent_conditional_implementation(
     Multi-output GP with independent GP priors.
 
     Number of latent processes equals the number of outputs (L = P).
-    The covariance matrices used to calculate the conditional have the following shape:
-
-    - Kuu: [P, M, M]
-    - Kuf: [P, M, N]
-    - Kff: [P, N] or [P, N, N]
 
     Further reference:
 
@@ -627,14 +588,17 @@ def separate_independent_conditional_implementation(
     fs = tf.transpose(f)[:, :, None]  # [P, M, 1]
     # [P, 1, M, M]  or  [P, M, 1]
 
-    base_conditional_args_to_map: Tuple[tf.Tensor, ...]
-    single_gp_conditional: Callable[[Tuple[tf.Tensor, ...]], MeanAndVariance]
-
     if q_sqrt is not None:
         q_sqrts = (
             tf.transpose(q_sqrt)[:, :, None] if q_sqrt.shape.ndims == 2 else q_sqrt[:, None, :, :]
         )
-        base_conditional_args_to_map = (Kmms, Kmns, Knns, fs, q_sqrts)
+        base_conditional_args_to_map = (
+            Kmms,
+            Kmns,
+            Knns,
+            fs,
+            q_sqrts,
+        )  # type: Tuple[tf.Tensor, ...]
 
         def single_gp_conditional(
             t: Tuple[tf.Tensor, ...]
