@@ -17,6 +17,7 @@ from typing import Optional
 import tensorflow as tf
 
 from ..base import Parameter, TensorType
+from ..experimental.check_shapes import check_shapes, inherit_check_shapes
 from ..utilities import positive
 from .base import ActiveDims, Kernel
 
@@ -31,6 +32,9 @@ class Linear(Kernel):
     where σ² is the variance parameter.
     """
 
+    @check_shapes(
+        "variance: [broadcast n_active_dims]",
+    )
     def __init__(
         self, variance: TensorType = 1.0, active_dims: Optional[ActiveDims] = None
     ) -> None:
@@ -52,12 +56,14 @@ class Linear(Kernel):
         ndims: int = self.variance.shape.ndims
         return ndims > 0
 
+    @inherit_check_shapes
     def K(self, X: TensorType, X2: Optional[TensorType] = None) -> tf.Tensor:
         if X2 is None:
             return tf.matmul(X * self.variance, X, transpose_b=True)
         else:
             return tf.tensordot(X * self.variance, X2, [[-1], [-1]])
 
+    @inherit_check_shapes
     def K_diag(self, X: TensorType) -> tf.Tensor:
         return tf.reduce_sum(tf.square(X) * self.variance, axis=-1)
 
@@ -75,6 +81,9 @@ class Polynomial(Linear):
     d is the degree parameter.
     """
 
+    @check_shapes(
+        "variance: [broadcast n_active_dims]",
+    )
     def __init__(
         self,
         degree: TensorType = 3.0,
@@ -94,8 +103,10 @@ class Polynomial(Linear):
         self.degree = degree
         self.offset = Parameter(offset, transform=positive())
 
+    @inherit_check_shapes
     def K(self, X: TensorType, X2: Optional[TensorType] = None) -> tf.Tensor:
         return (super().K(X, X2) + self.offset) ** self.degree
 
+    @inherit_check_shapes
     def K_diag(self, X: TensorType) -> tf.Tensor:
         return (super().K_diag(X) + self.offset) ** self.degree
