@@ -20,6 +20,7 @@ import tensorflow as tf
 from .. import logdensities
 from ..base import AnyNDArray, MeanAndVariance, Parameter, TensorType
 from ..config import default_float
+from ..experimental.check_shapes import check_shapes, inherit_check_shapes
 from ..utilities import positive, to_default_int
 from .base import ScalarLikelihood
 from .utils import inv_probit
@@ -50,17 +51,21 @@ class Poisson(ScalarLikelihood):
         self.invlink = invlink
         self.binsize: AnyNDArray = np.array(binsize, dtype=default_float())
 
-    def _scalar_log_prob(self, F: TensorType, Y: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _scalar_log_prob(self, X: TensorType, F: TensorType, Y: TensorType) -> tf.Tensor:
         return logdensities.poisson(Y, self.invlink(F) * self.binsize)
 
-    def _conditional_variance(self, F: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _conditional_variance(self, X: TensorType, F: TensorType) -> tf.Tensor:
         return self.invlink(F) * self.binsize
 
-    def _conditional_mean(self, F: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _conditional_mean(self, X: TensorType, F: TensorType) -> tf.Tensor:
         return self.invlink(F) * self.binsize
 
+    @inherit_check_shapes
     def _variational_expectations(
-        self, Fmu: TensorType, Fvar: TensorType, Y: TensorType
+        self, X: TensorType, Fmu: TensorType, Fvar: TensorType, Y: TensorType
     ) -> tf.Tensor:
         if self.invlink is tf.exp:
             return tf.reduce_sum(
@@ -70,7 +75,7 @@ class Poisson(ScalarLikelihood):
                 + Y * tf.math.log(self.binsize),
                 axis=-1,
             )
-        return super()._variational_expectations(Fmu, Fvar, Y)
+        return super()._variational_expectations(X, Fmu, Fvar, Y)
 
 
 class Bernoulli(ScalarLikelihood):
@@ -80,26 +85,35 @@ class Bernoulli(ScalarLikelihood):
         super().__init__(**kwargs)
         self.invlink = invlink
 
-    def _scalar_log_prob(self, F: TensorType, Y: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _scalar_log_prob(self, X: TensorType, F: TensorType, Y: TensorType) -> tf.Tensor:
         return logdensities.bernoulli(Y, self.invlink(F))
 
-    def _predict_mean_and_var(self, Fmu: TensorType, Fvar: TensorType) -> MeanAndVariance:
+    @inherit_check_shapes
+    def _predict_mean_and_var(
+        self, X: TensorType, Fmu: TensorType, Fvar: TensorType
+    ) -> MeanAndVariance:
         if self.invlink is inv_probit:
             p = inv_probit(Fmu / tf.sqrt(1 + Fvar))
             return p, p - tf.square(p)
         else:
             # for other invlink, use quadrature
-            return super()._predict_mean_and_var(Fmu, Fvar)
+            return super()._predict_mean_and_var(X, Fmu, Fvar)
 
-    def _predict_log_density(self, Fmu: TensorType, Fvar: TensorType, Y: TensorType) -> tf.Tensor:
-        p = self.predict_mean_and_var(Fmu, Fvar)[0]
+    @inherit_check_shapes
+    def _predict_log_density(
+        self, X: TensorType, Fmu: TensorType, Fvar: TensorType, Y: TensorType
+    ) -> tf.Tensor:
+        p = self.predict_mean_and_var(X, Fmu, Fvar)[0]
         return tf.reduce_sum(logdensities.bernoulli(Y, p), axis=-1)
 
-    def _conditional_mean(self, F: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _conditional_mean(self, X: TensorType, F: TensorType) -> tf.Tensor:
         return self.invlink(F)
 
-    def _conditional_variance(self, F: TensorType) -> tf.Tensor:
-        p = self.conditional_mean(F)
+    @inherit_check_shapes
+    def _conditional_variance(self, X: TensorType, F: TensorType) -> tf.Tensor:
+        p = self.conditional_mean(X, F)
         return p - (p ** 2)
 
 
@@ -123,6 +137,9 @@ class Ordinal(ScalarLikelihood):
     A reference is :cite:t:`chu2005gaussian`.
     """
 
+    @check_shapes(
+        "bin_edges: [num_bins_minus_1]",
+    )
     def __init__(self, bin_edges: AnyNDArray, **kwargs: Any) -> None:
         """
         bin_edges is a numpy array specifying at which function value the
@@ -134,7 +151,8 @@ class Ordinal(ScalarLikelihood):
         self.num_bins = bin_edges.size + 1
         self.sigma = Parameter(1.0, transform=positive())
 
-    def _scalar_log_prob(self, F: TensorType, Y: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _scalar_log_prob(self, X: TensorType, F: TensorType, Y: TensorType) -> tf.Tensor:
         Y = to_default_int(Y)
         scaled_bins_left = tf.concat([self.bin_edges / self.sigma, np.array([np.inf])], 0)
         scaled_bins_right = tf.concat([np.array([-np.inf]), self.bin_edges / self.sigma], 0)
@@ -147,6 +165,10 @@ class Ordinal(ScalarLikelihood):
             + 1e-6
         )
 
+    @check_shapes(
+        "F: [batch..., latent_dim]",
+        "return: [batch_and_latent_dim, num_bins]",
+    )
     def _make_phi(self, F: TensorType) -> tf.Tensor:
         """
         A helper function for making predictions. Constructs a probability
@@ -161,12 +183,14 @@ class Ordinal(ScalarLikelihood):
             scaled_bins_right - tf.reshape(F, (-1, 1)) / self.sigma
         )
 
-    def _conditional_mean(self, F: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _conditional_mean(self, X: TensorType, F: TensorType) -> tf.Tensor:
         phi = self._make_phi(F)
         Ys = tf.reshape(np.arange(self.num_bins, dtype=default_float()), (-1, 1))
         return tf.reshape(tf.linalg.matmul(phi, Ys), tf.shape(F))
 
-    def _conditional_variance(self, F: TensorType) -> tf.Tensor:
+    @inherit_check_shapes
+    def _conditional_variance(self, X: TensorType, F: TensorType) -> tf.Tensor:
         phi = self._make_phi(F)
         Ys = tf.reshape(np.arange(self.num_bins, dtype=default_float()), (-1, 1))
         E_y = phi @ Ys
