@@ -15,12 +15,12 @@
 import enum
 from abc import ABC, abstractmethod
 from dataclasses import dataclass
-from typing import Optional, Tuple, Type, Union
+from typing import Optional, Tuple, Type, Union, cast
 
 import tensorflow as tf
 
 from . import covariances, kernels, mean_functions
-from .base import MeanAndVariance, Module, Parameter, RegressionData, TensorType
+from .base import MeanAndVariance, Module, RegressionData, TensorType
 from .conditionals.util import (
     base_conditional,
     base_conditional_with_lm,
@@ -32,6 +32,14 @@ from .conditionals.util import (
 )
 from .config import default_float, default_jitter
 from .covariances import Kuf, Kuu
+from .experimental.check_shapes import (
+    ErrorContext,
+    Shape,
+    check_shapes,
+    get_shape,
+    inherit_check_shapes,
+    register_get_shape,
+)
 from .inducing_variables import (
     FallbackSeparateIndependentInducingVariables,
     FallbackSharedIndependentInducingVariables,
@@ -41,8 +49,9 @@ from .inducing_variables import (
     SharedIndependentInducingVariables,
 )
 from .kernels import Kernel
+from .likelihoods import Gaussian
 from .mean_functions import MeanFunction
-from .utilities import Dispatcher, add_noise_cov
+from .utilities import Dispatcher, add_likelihood_noise_cov, assert_params_false
 from .utilities.ops import eye, leading_transpose
 
 
@@ -54,8 +63,11 @@ class _QDistribution(Module):
 
 
 class _DeltaDist(_QDistribution):
+    @check_shapes(
+        "q_mu: [M, L]",
+    )
     def __init__(self, q_mu: TensorType) -> None:
-        self.q_mu = q_mu  # [M, L]
+        self.q_mu = q_mu
 
     @property
     def q_sqrt(self) -> Optional[tf.Tensor]:
@@ -63,15 +75,23 @@ class _DeltaDist(_QDistribution):
 
 
 class _DiagNormal(_QDistribution):
+    @check_shapes(
+        "q_mu: [M, L]",
+        "q_sqrt: [M, L]",
+    )
     def __init__(self, q_mu: TensorType, q_sqrt: TensorType) -> None:
-        self.q_mu = q_mu  # [M, L]
-        self.q_sqrt = q_sqrt  # [M, L]
+        self.q_mu = q_mu
+        self.q_sqrt = q_sqrt
 
 
 class _MvNormal(_QDistribution):
+    @check_shapes(
+        "q_mu: [M, L]",
+        "q_sqrt: [L, M, M]  # lower-triangular",
+    )
     def __init__(self, q_mu: TensorType, q_sqrt: TensorType) -> None:
-        self.q_mu = q_mu  # [M, L]
-        self.q_sqrt = q_sqrt  # [L, M, M], lower-triangular
+        self.q_mu = q_mu
+        self.q_sqrt = q_sqrt
 
 
 class PrecomputeCacheType(enum.Enum):
@@ -115,6 +135,10 @@ class PrecomputedValue:
         )
 
     @staticmethod
+    @check_shapes(
+        "alpha: [M_L_or_L_M_M...]",
+        "Qinv: [M_M_or_L_M_M...]",
+    )
     def wrap_alpha_Qinv(alpha: TensorType, Qinv: TensorType) -> Tuple["PrecomputedValue", ...]:
         """
         Wraps `alpha` and `Qinv` in `PrecomputedValue`\ s.
@@ -145,6 +169,11 @@ class PrecomputedValue:
         )
 
 
+@register_get_shape(PrecomputedValue)
+def get_precomputed_value_shape(shaped: PrecomputedValue, context: ErrorContext) -> Shape:
+    return get_shape(shaped.value, context)
+
+
 def _validate_precompute_cache_type(
     value: Union[None, PrecomputeCacheType, str]
 ) -> PrecomputeCacheType:
@@ -162,6 +191,9 @@ def _validate_precompute_cache_type(
 
 
 class AbstractPosterior(Module, ABC):
+    @check_shapes(
+        "X_data: [N_D_or_M_D_P...]",
+    )
     def __init__(
         self,
         kernel: Kernel,
@@ -185,6 +217,11 @@ class AbstractPosterior(Module, ABC):
 
         self._precompute_cache: Optional[PrecomputeCacheType] = None
 
+    @check_shapes(
+        "Xnew: [batch..., D]",
+        "mean: [batch..., Q]",
+        "return: [batch..., Q]",
+    )
     def _add_mean_function(self, Xnew: TensorType, mean: TensorType) -> tf.Tensor:
         if self.mean_function is None:
             return mean
@@ -200,6 +237,14 @@ class AbstractPosterior(Module, ABC):
         `cache` argument.
         """
 
+    @check_shapes(
+        "Xnew: [batch..., N, D]",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def fused_predict_f(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -213,6 +258,14 @@ class AbstractPosterior(Module, ABC):
         return self._add_mean_function(Xnew, mean), cov
 
     @abstractmethod
+    @check_shapes(
+        "Xnew: [batch..., N, D]",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -221,6 +274,14 @@ class AbstractPosterior(Module, ABC):
         Does not make use of caching
         """
 
+    @check_shapes(
+        "Xnew: [batch..., N, D]",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def predict_f(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -238,6 +299,14 @@ class AbstractPosterior(Module, ABC):
         return self._add_mean_function(Xnew, mean), cov
 
     @abstractmethod
+    @check_shapes(
+        "Xnew: [batch..., N, D]",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -290,11 +359,15 @@ class AbstractPosterior(Module, ABC):
 
 
 class GPRPosterior(AbstractPosterior):
+    @check_shapes(
+        "data[0]: [N, D]",
+        "data[1]: [N, Q]",
+    )
     def __init__(
         self,
         kernel: Kernel,
         data: RegressionData,
-        likelihood_variance: Parameter,
+        likelihood: Gaussian,
         mean_function: MeanFunction,
         *,
         precompute_cache: Optional[PrecomputeCacheType],
@@ -302,11 +375,12 @@ class GPRPosterior(AbstractPosterior):
         X, Y = data
         super().__init__(kernel, X, mean_function=mean_function)
         self.Y_data = Y
-        self.likelihood_variance = likelihood_variance
+        self.likelihood = likelihood
 
         if precompute_cache is not None:
             self.update_cache(precompute_cache)
 
+    @inherit_check_shapes
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -318,70 +392,46 @@ class GPRPosterior(AbstractPosterior):
         Computes predictive mean and (co)variance at Xnew, *excluding* mean_function.
         Relies on cached alpha and Qinv.
         """
-        (alpha,) = cache
-        (Qinv,) = cache
+        assert_params_false(self._conditional_with_precompute, full_output_cov=full_output_cov)
+        err, Lm = cache
 
-        Kmn = self.kernel(self.X_data, Xnew)
-        # compute kernel stuff
-        num_func = tf.shape(self.Y_data)[-1]  # R
-        N = tf.shape(Kmn)[-1]
-
-        # get the leading dims in Kmn to the front of the tensor Kmn
-        K = tf.rank(Kmn)
-        perm = tf.concat(
-            [
-                tf.reshape(tf.range(1, K - 1), [K - 2]),  # leading dims (...)
-                tf.reshape(0, [1]),  # [M]
-                tf.reshape(K - 1, [1]),
-            ],
-            0,
-        )  # [N]
-        Kmn = tf.transpose(Kmn, perm)  # [..., M, N]
-        leading_dims = tf.shape(Kmn)[:-2]
-
-        # get the leading dims in Knm to the front of the tensor Knm
-        Knm = leading_transpose(Kmn, [..., -1, -2])
-
-        assert self.mean_function is not None
         Knn = self.kernel(Xnew, full_cov=full_cov)
-        err = self.Y_data - self.mean_function(self.X_data)
+        Kmn = self.kernel(self.X_data, Xnew)
 
-        mean = Knm @ alpha @ err
-
-        # The GPR model only has a single latent GP.
-        if full_cov:
-            cov = Knn - Knm @ Qinv @ Kmn  # [..., N, N]
-            cov_shape = tf.concat([leading_dims, [num_func, N, N]], 0)
-            cov = tf.broadcast_to(tf.expand_dims(cov, -3), cov_shape)  # [..., R, N, N]
-
-        else:
-            cov = Knn - tf.einsum("...ij,...ji->...i", Knm @ Qinv, Kmn)  # [..., N]
-            cov_shape = tf.concat([leading_dims, [num_func, N]], 0)  # [..., R, N]
-            cov = tf.broadcast_to(tf.expand_dims(cov, -2), cov_shape)  # [..., R, N]
-            cov = tf.linalg.adjoint(cov)
-
-        return mean, cov
-
-    def _precompute(self) -> Tuple[PrecomputedValue, ...]:
-        Kmm = self.kernel(self.X_data)
-        Kmm_plus_s = add_noise_cov(Kmm, self.likelihood_variance)
-
-        Lm = tf.linalg.cholesky(Kmm_plus_s)
-        Kmm_plus_s_inv = tf.linalg.cholesky_solve(
-            Lm, tf.eye(tf.shape(self.X_data)[0], dtype=Lm.dtype)
+        return base_conditional_with_lm(
+            Kmn=Kmn,
+            Lm=Lm,
+            Knn=Knn,
+            f=err,
+            full_cov=full_cov,
+            q_sqrt=None,
+            white=False,
         )
 
-        M = self.X_data.shape[0]
+    @check_shapes(
+        "return[0]: [M, D]",
+        "return[1]: [M, M]",
+    )
+    def _precompute(self) -> Tuple[PrecomputedValue, ...]:
+        assert self.mean_function is not None
+        X_data = cast(tf.Tensor, self.X_data)
+        err = self.Y_data - self.mean_function(X_data)
+
+        Kmm = self.kernel(X_data)
+        Kmm_plus_s = add_likelihood_noise_cov(Kmm, self.likelihood, X_data)
+        Lm = tf.linalg.cholesky(Kmm_plus_s)
+
+        D = err.shape[1]
+        M = X_data.shape[0]
+        D_dynamic = D is None
         M_dynamic = M is None
 
-        tf.debugging.assert_shapes(
-            [
-                (Kmm_plus_s_inv, ["M", "M"]),
-                (Kmm, ["M", "M"]),
-            ]
+        return (
+            PrecomputedValue(err, (M_dynamic, D_dynamic)),
+            PrecomputedValue(Lm, (M_dynamic, M_dynamic)),
         )
-        return (PrecomputedValue(Kmm_plus_s_inv, (M_dynamic, M_dynamic)),)
 
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -389,19 +439,8 @@ class GPRPosterior(AbstractPosterior):
         Computes predictive mean and (co)variance at Xnew, *excluding* mean_function
         Does not make use of caching
         """
-
-        # taken directly from the deprecated GPR implementation
-        assert self.mean_function is not None
-        err = self.Y_data - self.mean_function(self.X_data)
-
-        Kmm = self.kernel(self.X_data)
-        Knn = self.kernel(Xnew, full_cov=full_cov)
-        Kmn = self.kernel(self.X_data, Xnew)
-        Kmm_plus_s = add_noise_cov(Kmm, self.likelihood_variance)
-
-        return base_conditional(
-            Kmn, Kmm_plus_s, Knn, err, full_cov=full_cov, white=False
-        )  # [N, P], [N, P] or [P, N, N]
+        temp_cache = tuple(c.value for c in self._precompute())
+        return self._conditional_with_precompute(temp_cache, Xnew, full_cov, full_output_cov)
 
 
 class SGPRPosterior(AbstractPosterior):
@@ -410,12 +449,17 @@ class SGPRPosterior(AbstractPosterior):
     models to compute faster predictions on unseen points.
     """
 
+    @check_shapes(
+        "data[0]: [N, D]",
+        "data[1]: [N, Q]",
+        "inducing_variable: [M, D, 1]",
+    )
     def __init__(
         self,
         kernel: Kernel,
         data: RegressionData,
         inducing_variable: InducingPoints,
-        likelihood_variance: Parameter,
+        likelihood: Gaussian,
         num_latent_gps: int,
         mean_function: MeanFunction,
         *,
@@ -424,43 +468,14 @@ class SGPRPosterior(AbstractPosterior):
         X, Y = data
         super().__init__(kernel, X, mean_function=mean_function)
         self.Y_data = Y
-        self.likelihood_variance = likelihood_variance
+        self.likelihood = likelihood
         self.inducing_variable = inducing_variable
         self.num_latent_gps = num_latent_gps
 
         if precompute_cache is not None:
             self.update_cache(precompute_cache)
 
-    def _precompute(self) -> Tuple[PrecomputedValue, ...]:
-        # taken directly from the deprecated SGPR implementation
-        num_inducing = self.inducing_variable.num_inducing
-        assert self.mean_function is not None
-        err = self.Y_data - self.mean_function(self.X_data)
-        kuf = Kuf(self.inducing_variable, self.kernel, self.X_data)
-        kuu = Kuu(self.inducing_variable, self.kernel, jitter=default_jitter())
-        sigma = tf.sqrt(self.likelihood_variance)
-        L = tf.linalg.cholesky(kuu)  # cache alpha, qinv
-        A = tf.linalg.triangular_solve(L, kuf, lower=True) / sigma
-        B = tf.linalg.matmul(A, A, transpose_b=True) + tf.eye(
-            num_inducing, dtype=default_float()
-        )  # cache qinv
-        LB = tf.linalg.cholesky(B)  # cache alpha
-        Aerr = tf.linalg.matmul(A, err)
-        c = tf.linalg.triangular_solve(LB, Aerr, lower=True) / sigma  # cache alpha
-
-        # get intermediate variables
-        Linv = tf.linalg.triangular_solve(L, tf.eye(num_inducing, dtype=default_float()))
-        LBinv = tf.linalg.triangular_solve(LB, tf.eye(num_inducing, dtype=default_float()))
-        Binv = tf.linalg.inv(B)  # naive...can do better?
-        tmp = tf.eye(num_inducing, dtype=default_float()) - Binv
-
-        # calculate cached values
-        LinvT = tf.transpose(Linv)
-        alpha = LinvT @ tf.transpose(LBinv) @ c
-        Qinv = LinvT @ tmp @ Linv
-
-        return PrecomputedValue.wrap_alpha_Qinv(alpha, Qinv)
-
+    @inherit_check_shapes
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -472,48 +487,11 @@ class SGPRPosterior(AbstractPosterior):
         Computes predictive mean and (co)variance at Xnew, *excluding* mean_function.
         Relies on cached alpha and Qinv.
         """
-        alpha, Qinv = cache
+        assert_params_false(self._conditional_with_precompute, full_output_cov=full_output_cov)
+
+        L, LB, c = cache
 
         Kus = Kuf(self.inducing_variable, self.kernel, Xnew)
-        Knn = self.kernel(Xnew, full_cov=full_cov)
-
-        Ksu = tf.transpose(Kus)
-        mean = Ksu @ alpha
-
-        if full_cov:
-            var = Knn - Ksu @ Qinv @ Kus
-            var = tf.tile(var[None, ...], [self.num_latent_gps, 1, 1])  # [P, N, N]
-        else:
-            Kfu_Qinv_Kuf = tf.reduce_sum(Kus * tf.matmul(Qinv, Kus), axis=-2)
-            var = Knn - Kfu_Qinv_Kuf
-            var = tf.tile(var[:, None], [1, self.num_latent_gps])
-
-        return mean, var
-
-    def _conditional_fused(
-        self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
-    ) -> MeanAndVariance:
-        """
-        Compute the mean and variance of the latent function at some new points
-        Xnew. Does not make use of caching
-        """
-
-        # taken directly from the deprecated SGPR implementation
-        num_inducing = self.inducing_variable.num_inducing
-        assert self.mean_function is not None
-        err = self.Y_data - self.mean_function(self.X_data)
-        kuf = Kuf(self.inducing_variable, self.kernel, self.X_data)
-        kuu = Kuu(self.inducing_variable, self.kernel, jitter=default_jitter())
-        Kus = Kuf(self.inducing_variable, self.kernel, Xnew)
-        sigma = tf.sqrt(self.likelihood_variance)
-        L = tf.linalg.cholesky(kuu)  # cache alpha, qinv
-        A = tf.linalg.triangular_solve(L, kuf, lower=True) / sigma
-        B = tf.linalg.matmul(A, A, transpose_b=True) + tf.eye(
-            num_inducing, dtype=default_float()
-        )  # cache qinv
-        LB = tf.linalg.cholesky(B)  # cache alpha
-        Aerr = tf.linalg.matmul(A, err)
-        c = tf.linalg.triangular_solve(LB, Aerr, lower=True) / sigma  # cache alpha
         tmp1 = tf.linalg.triangular_solve(L, Kus, lower=True)
         tmp2 = tf.linalg.triangular_solve(LB, tmp1, lower=True)
         mean = tf.linalg.matmul(tmp2, c, transpose_a=True)
@@ -534,8 +512,62 @@ class SGPRPosterior(AbstractPosterior):
 
         return mean, var
 
+    @check_shapes(
+        "return[0]: [M, M]",
+        "return[1]: [M, M]",
+        "return[2]: [M, D]",
+    )
+    def _precompute(self) -> Tuple[PrecomputedValue, ...]:
+        assert self.mean_function is not None
+
+        X_data = cast(tf.Tensor, self.X_data)
+        num_inducing = self.inducing_variable.num_inducing
+        err = self.Y_data - self.mean_function(X_data)
+
+        kuf = Kuf(self.inducing_variable, self.kernel, X_data)
+        kuu = Kuu(self.inducing_variable, self.kernel, jitter=default_jitter())
+
+        sigma_sq = tf.squeeze(self.likelihood.variance_at(X_data), axis=-1)
+        sigma = tf.sqrt(sigma_sq)
+
+        L = tf.linalg.cholesky(kuu)  # cache alpha, qinv
+        A = tf.linalg.triangular_solve(L, kuf / sigma, lower=True)
+        B = tf.linalg.matmul(A, A, transpose_b=True) + tf.eye(
+            num_inducing, dtype=default_float()
+        )  # cache qinv
+        LB = tf.linalg.cholesky(B)  # cache alpha
+        Aerr = tf.linalg.matmul(A, err / sigma[..., None])
+        c = tf.linalg.triangular_solve(LB, Aerr, lower=True)
+
+        D = err.shape[1]
+        M = X_data.shape[0]
+        D_dynamic = D is None
+        M_dynamic = M is None
+
+        return (
+            PrecomputedValue(L, (M_dynamic, M_dynamic)),
+            PrecomputedValue(LB, (M_dynamic, M_dynamic)),
+            PrecomputedValue(c, (M_dynamic, D_dynamic)),
+        )
+
+    @inherit_check_shapes
+    def _conditional_fused(
+        self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
+    ) -> MeanAndVariance:
+        """
+        Compute the mean and variance of the latent function at some new points
+        Xnew. Does not make use of caching
+        """
+        temp_cache = tuple(c.value for c in self._precompute())
+        return self._conditional_with_precompute(temp_cache, Xnew, full_cov, full_output_cov)
+
 
 class VGPPosterior(AbstractPosterior):
+    @check_shapes(
+        "X: [N, D]",
+        "q_mu: [N, P]",
+        "q_sqrt: [N_P_or_P_N_N...]",
+    )
     def __init__(
         self,
         kernel: Kernel,
@@ -555,6 +587,7 @@ class VGPPosterior(AbstractPosterior):
         if precompute_cache is not None:
             self.update_cache(precompute_cache)
 
+    @inherit_check_shapes
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -562,6 +595,8 @@ class VGPPosterior(AbstractPosterior):
         full_cov: bool = False,
         full_output_cov: bool = False,
     ) -> MeanAndVariance:
+        assert_params_false(self._conditional_with_precompute, full_output_cov=full_output_cov)
+
         (Lm,) = cache
         Kmn = self.kernel(self.X_data, Xnew)  # [M, ..., N]
         Knn = self.kernel(
@@ -578,17 +613,22 @@ class VGPPosterior(AbstractPosterior):
             white=self.white,
         )
 
+    @check_shapes(
+        "return[0]: [M, M]",
+    )
     def _precompute(self) -> Tuple[PrecomputedValue, ...]:
-        Kmm = self.kernel(self.X_data) + eye(
-            tf.shape(self.X_data)[-2], value=default_jitter(), dtype=self.X_data.dtype
+        X_data = cast(tf.Tensor, self.X_data)
+        Kmm = self.kernel(X_data) + eye(
+            tf.shape(X_data)[-2], value=default_jitter(), dtype=X_data.dtype
         )  # [..., M, M]
         Lm = tf.linalg.cholesky(Kmm)
 
-        M = self.X_data.shape[0]
+        M = X_data.shape[0]
         M_dynamic = M is None
 
         return (PrecomputedValue(Lm, (M_dynamic, M_dynamic)),)
 
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -597,6 +637,11 @@ class VGPPosterior(AbstractPosterior):
 
 
 class BasePosterior(AbstractPosterior):
+    @check_shapes(
+        "inducing_variable: [M, D, broadcast P]",
+        "q_mu: [N, P]",
+        "q_sqrt: [N_P_or_P_N_N...]",
+    )
     def __init__(
         self,
         kernel: Kernel,
@@ -616,15 +661,25 @@ class BasePosterior(AbstractPosterior):
         if precompute_cache is not None:
             self.update_cache(precompute_cache)
 
-    @property
+    @property  # type: ignore[misc]
+    @check_shapes(
+        "return: [N, P]",
+    )
     def q_mu(self) -> tf.Tensor:
         return self._q_dist.q_mu
 
-    @property
+    @property  # type: ignore[misc]
+    @check_shapes(
+        "return: [N_P_or_P_N_N...]",
+    )
     def q_sqrt(self) -> tf.Tensor:
         return self._q_dist.q_sqrt
 
-    def _set_qdist(self, q_mu: TensorType, q_sqrt: TensorType) -> tf.Tensor:
+    @check_shapes(
+        "q_mu: [N, P]",
+        "q_sqrt: [N_P_or_P_N_N...]",
+    )
+    def _set_qdist(self, q_mu: TensorType, q_sqrt: TensorType) -> None:
         if q_sqrt is None:
             self._q_dist = _DeltaDist(q_mu)
         elif len(q_sqrt.shape) == 2:  # q_diag
@@ -632,6 +687,10 @@ class BasePosterior(AbstractPosterior):
         else:
             self._q_dist = _MvNormal(q_mu, q_sqrt)
 
+    @check_shapes(
+        "return[0]: [M_L_or_L_M_M...]",
+        "return[1]: [L, M, M]",
+    )
     def _precompute(self) -> Tuple[PrecomputedValue, ...]:
         Kuu = covariances.Kuu(self.X_data, self.kernel, jitter=default_jitter())  # [(R), M, M]
         q_mu = self._q_dist.q_mu
@@ -684,21 +743,30 @@ class BasePosterior(AbstractPosterior):
         M, L = tf.unstack(tf.shape(self._q_dist.q_mu), num=2)
         Qinv = tf.broadcast_to(Qinv, [L, M, M])
 
-        tf.debugging.assert_shapes(
-            [
-                (Qinv, ["L", "M", "M"]),
-            ]
-        )
-
         return PrecomputedValue.wrap_alpha_Qinv(alpha, Qinv)
 
 
 class IndependentPosterior(BasePosterior):
+    @check_shapes(
+        "mean: [batch..., N, P]",
+        "cov: [batch..., P, N, N] if full_cov",
+        "cov: [batch..., N, P] if not full_cov",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def _post_process_mean_and_cov(
         self, mean: TensorType, cov: TensorType, full_cov: bool, full_output_cov: bool
     ) -> MeanAndVariance:
         return mean, expand_independent_outputs(cov, full_cov, full_output_cov)
 
+    @check_shapes(
+        "Xnew: [N, D]",
+        "return: [broadcast P, N, N] if full_cov",
+        "return: [broadcast P, N] if (not full_cov)",
+    )
     def _get_Kff(self, Xnew: TensorType, full_cov: bool) -> tf.Tensor:
 
         # TODO: this assumes that Xnew has shape [N, D] and no leading dims
@@ -722,6 +790,7 @@ class IndependentPosterior(BasePosterior):
 
         return Kff
 
+    @inherit_check_shapes
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -755,6 +824,7 @@ class IndependentPosterior(BasePosterior):
 
 class IndependentPosteriorSingleOutput(IndependentPosterior):
     # could almost be the same as IndependentPosteriorMultiOutput ...
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -772,6 +842,7 @@ class IndependentPosteriorSingleOutput(IndependentPosterior):
 
 
 class IndependentPosteriorMultiOutput(IndependentPosterior):
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -817,19 +888,26 @@ class IndependentPosteriorMultiOutput(IndependentPosterior):
 
 
 class LinearCoregionalizationPosterior(IndependentPosteriorMultiOutput):
+    @check_shapes(
+        "mean: [batch..., N, L]",
+        "cov: [batch..., L, N, N] if full_cov",
+        "cov: [batch..., N, L] if not full_cov",
+        "return[0]: [batch..., N, P]",
+        "return[1]: [batch..., N, P, N, P] if full_cov and full_output_cov",
+        "return[1]: [batch..., N, P, P] if (not full_cov) and full_output_cov",
+        "return[1]: [batch..., P, N, N] if full_cov and (not full_output_cov)",
+        "return[1]: [batch..., N, P] if (not full_cov) and (not full_output_cov)",
+    )
     def _post_process_mean_and_cov(
         self, mean: TensorType, cov: TensorType, full_cov: bool, full_output_cov: bool
     ) -> MeanAndVariance:
-        """
-        mean: [N, L]
-        cov: [L, N, N] or [N, L]
-        """
         cov = expand_independent_outputs(cov, full_cov, full_output_cov=False)
         mean, cov = mix_latent_gp(self.kernel.W, mean, cov, full_cov, full_output_cov)
         return mean, cov
 
 
 class FullyCorrelatedPosterior(BasePosterior):
+    @inherit_check_shapes
     def _conditional_with_precompute(
         self,
         cache: Tuple[tf.Tensor, ...],
@@ -897,6 +975,7 @@ class FullyCorrelatedPosterior(BasePosterior):
 
         return mean, cov
 
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -934,6 +1013,7 @@ class FullyCorrelatedPosterior(BasePosterior):
 
 
 class FallbackIndependentLatentPosterior(FullyCorrelatedPosterior):  # XXX
+    @inherit_check_shapes
     def _conditional_fused(
         self, Xnew: TensorType, full_cov: bool = False, full_output_cov: bool = False
     ) -> MeanAndVariance:
@@ -1017,7 +1097,7 @@ def create_posterior(
 ) -> BasePosterior:
     posterior_class = get_posterior_class(kernel, inducing_variable)
     precompute_cache = _validate_precompute_cache_type(precompute_cache)
-    return posterior_class(  # type: ignore
+    return posterior_class(  # type: ignore[no-any-return]
         kernel,
         inducing_variable,
         q_mu,
