@@ -19,6 +19,7 @@ import tensorflow_probability as tfp
 
 from ..base import TensorData
 from ..config import default_float, default_int
+from ..experimental.check_shapes import check_shapes
 from .ops import cast
 
 __all__ = [
@@ -30,10 +31,18 @@ __all__ = [
 ]
 
 
+@check_shapes(
+    "x: [any...]",
+    "return: [any...]",
+)
 def to_default_int(x: TensorData) -> tf.Tensor:
     return cast(x, dtype=default_int())
 
 
+@check_shapes(
+    "x: [any...]",
+    "return: [any...]",
+)
 def to_default_float(x: TensorData) -> tf.Tensor:
     return cast(x, dtype=default_float())
 
