@@ -475,7 +475,7 @@ def test_check_shapes__invalid_argument() -> None:
     with pytest.raises(TypeError):
         # Linter disables, because we're intentionally making an invalid call.
         # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
-        f(c=t(2, 3))  # type: ignore
+        f(c=t(2, 3))  # type: ignore[call-arg]
 
 
 def test_check_shapes__argument_refs() -> None:
@@ -564,15 +564,18 @@ def test_check_shapes__condition() -> None:
     @check_shapes(
         "a: [1] if b1",
         "a: [2] if not b1 and b2",
-        "a: [3] if not b1 and not b2",
+        "a: [3] if not b1 and b2 is not None and not b2",
+        "a: [4] if not b1 and b2 is None",
     )
-    def f(a: TestShaped, b1: bool, b2: bool) -> None:
+    def f(a: TestShaped, b1: bool, b2: Optional[bool]) -> None:
         pass
 
     f(t(1), True, True)
     f(t(1), True, False)
+    f(t(1), True, None)
     f(t(2), False, True)
     f(t(3), False, False)
+    f(t(4), False, None)
 
     with pytest.raises(ShapeMismatchError):
         f(t(2), True, True)
@@ -580,12 +583,18 @@ def test_check_shapes__condition() -> None:
     with pytest.raises(ShapeMismatchError):
         f(t(2), True, False)
 
+    with pytest.raises(ShapeMismatchError):
+        f(t(2), True, None)
+
     with pytest.raises(ShapeMismatchError):
         f(t(1), False, True)
 
     with pytest.raises(ShapeMismatchError):
         f(t(2), False, False)
 
+    with pytest.raises(ShapeMismatchError):
+        f(t(2), False, None)
+
 
 def test_check_shapes__reuse() -> None:
     check_my_shapes = check_shapes(
