@@ -17,6 +17,7 @@ from unittest.mock import MagicMock
 import pytest
 from lark.exceptions import UnexpectedCharacters, UnexpectedEOF, UnexpectedToken
 
+from gpflow.experimental.check_shapes import set_enable_function_call_precompute
 from gpflow.experimental.check_shapes.error_contexts import (
     ArgumentContext,
     AttributeContext,
@@ -37,7 +38,6 @@ from gpflow.experimental.check_shapes.error_contexts import (
     ShapeContext,
     StackContext,
     TensorSpecContext,
-    TrailingBroadcastVarrankContext,
     VariableContext,
 )
 from gpflow.experimental.check_shapes.specs import ParsedNoteSpec
@@ -52,6 +52,14 @@ def to_str(context: ErrorContext) -> str:
     return builder.build()
 
 
+def check_eq(context: ErrorContext) -> None:
+    # Sanity checking of __eq__ and __hash__
+    assert context == context  # pylint: disable=comparison-with-itself
+    assert context != TestContext()
+    assert context != None  # pylint: disable=singleton-comparison
+    assert context in {context}
+
+
 def test_message_builder__add_line() -> None:
     builder = MessageBuilder()
     builder.add_line("foo")
@@ -215,6 +223,7 @@ A
 )
 def test_stack_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -306,11 +315,14 @@ A
 )
 def test_parallel_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 def test_function_call_context() -> None:
     def f() -> str:
-        return to_str(FunctionCallContext(f))
+        context = FunctionCallContext(f)
+        check_eq(context)
+        return to_str(context)
 
     call_line = current_line() + 1
     assert (
@@ -323,7 +335,9 @@ f called at: {__file__}:{call_line}
 
 def test_function_call_context__wrapping() -> None:
     def f() -> str:
-        return to_str(FunctionCallContext(f3))
+        context = FunctionCallContext(f3)
+        check_eq(context)
+        return to_str(context)
 
     @wraps(f)
     def f2() -> str:
@@ -342,7 +356,23 @@ f called at: {__file__}:{call_line}
     )
 
 
-def test_function_call_context__precompute() -> None:
+def test_function_call_context__precompute__disabled() -> None:
+    set_enable_function_call_precompute(False)
+
+    def f() -> ErrorContext:
+        return FunctionCallContext(f).precompute()
+
+    context = f()
+
+    assert f"""
+f called at: <Unknown file>:<Unknown line> (Disabled. Call gpflow.experimental.check_shapes.set_enable_function_call_precompute(True) to see this. (Slow.))
+""" == to_str(
+        context
+    )
+    check_eq(context)
+
+
+def test_function_call_context__precompute__enabled() -> None:
     def f() -> ErrorContext:
         return FunctionCallContext(f).precompute()
 
@@ -354,6 +384,7 @@ f called at: {__file__}:{call_line}
 """ == to_str(
         context
     )
+    check_eq(context)
 
 
 def test_function_definition_context() -> None:
@@ -362,37 +393,45 @@ def test_function_definition_context() -> None:
     def f() -> None:
         ...
 
+    context = FunctionDefinitionContext(f)
     assert f"""
 Function: test_function_definition_context.<locals>.f
   Declared: {__file__}:{def_line}
 """ == to_str(
-        FunctionDefinitionContext(f)
+        context
     )
+    check_eq(context)
 
 
 def test_function_definition_context__builtin() -> None:
+    context = FunctionDefinitionContext(str)
     assert """
 Function: str
   Declared: <Unknown file>:<Unknown line>
 """ == to_str(
-        FunctionDefinitionContext(str)
+        context
     )
+    check_eq(context)
 
 
 def test_variable_context() -> None:
+    context = VariableContext("foo")
     assert """
 Variable: foo
 """ == to_str(
-        VariableContext("foo")
+        context
     )
+    check_eq(context)
 
 
 def test_tensor_spec_context() -> None:
+    context = TensorSpecContext(make_tensor_spec(make_shape_spec("a", 1), None))
     assert """
 Specification: [a, 1]
 """ == to_str(
-        TensorSpecContext(make_tensor_spec(make_shape_spec("a", 1), None))
+        context
     )
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -415,6 +454,13 @@ Argument number (0-indexed): 3
             """
 Argument: arg
   Value: 7
+""",
+        ),
+        (
+            ArgumentContext("arg", value={}),
+            """
+Argument: arg
+  Value: {}
 """,
         ),
         (
@@ -428,6 +474,7 @@ Argument number (0-indexed): 3
 )
 def test_argument_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -444,6 +491,13 @@ Attribute: .attr
             """
 Attribute: .attr
   Value: 3
+""",
+        ),
+        (
+            AttributeContext("attr", value={}),
+            """
+Attribute: .attr
+  Value: {}
 """,
         ),
         (
@@ -457,6 +511,7 @@ Attribute: .attr
 )
 def test_attribute_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -473,6 +528,13 @@ Index: [2]
             """
 Index: [2]
   Value: 3
+""",
+        ),
+        (
+            IndexContext(2, value={}),
+            """
+Index: [2]
+  Value: {}
 """,
         ),
         (
@@ -486,6 +548,7 @@ Index: [2]
 )
 def test_index_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -501,6 +564,12 @@ Map key: [2]
             MappingKeyContext("foo"),
             """
 Map key: ['foo']
+""",
+        ),
+        (
+            MappingKeyContext({}),
+            """
+Map key: [{}]
 """,
         ),
         (
@@ -513,6 +582,7 @@ Map key: [None]
 )
 def test_mapping_key_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -529,6 +599,13 @@ Map value, of key: [2]
             """
 Map value, of key: ['foo']
   Value: 3
+""",
+        ),
+        (
+            MappingValueContext({}, value={}),
+            """
+Map value, of key: [{}]
+  Value: {}
 """,
         ),
         (
@@ -542,14 +619,17 @@ Map value, of key: [None]
 )
 def test_mapping_value_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 def test_condition_context() -> None:
+    context = ConditionContext(bor(barg("foo"), bnot(barg("bar"))))
     assert """
 Condition: foo or (not bar)
 """ == to_str(
-        ConditionContext(bor(barg("foo"), bnot(barg("bar"))))
+        context
     )
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -573,21 +653,24 @@ Actual:   [2, None]
             ShapeContext(make_shape_spec("x", 3), None),
             """
 Expected: [x, 3]
-Actual:   <Tensor has unknown shape>
+Actual:   <Tensor is None or has unknown shape>
 """,
         ),
     ],
 )
 def test_shape_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 def test_note_context() -> None:
+    context = NoteContext(ParsedNoteSpec("Some note."))
     assert """
 Note: Some note.
 """ == to_str(
-        NoteContext(ParsedNoteSpec("Some note."))
+        context
     )
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -603,6 +686,12 @@ Value: None
             ObjectValueContext(7),
             """
 Value: 7
+""",
+        ),
+        (
+            ObjectValueContext({}),
+            """
+Value: {}
 """,
         ),
         (
@@ -615,6 +704,7 @@ Value: 'foo'
 )
 def test_object_value_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -630,6 +720,12 @@ Object type: builtins.NoneType
             ObjectTypeContext(7),
             """
 Object type: builtins.int
+""",
+        ),
+        (
+            ObjectTypeContext({}),
+            """
+Object type: builtins.dict
 """,
         ),
         (
@@ -642,6 +738,7 @@ Object type: gpflow.experimental.check_shapes.error_contexts.ObjectTypeContext
 )
 def test_object_type_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
 
 
 def test_object_type_context__nested_type() -> None:
@@ -650,11 +747,13 @@ def test_object_type_context__nested_type() -> None:
 
             pass
 
+    context = ObjectTypeContext(A.B())
     assert """
 Object type: tests.gpflow.experimental.check_shapes.test_error_contexts.test_object_type_context__nested_type.<locals>.A.B
 """ == to_str(
-        ObjectTypeContext(A.B())
+        context
     )
+    check_eq(context)
 
 
 def test_lark_unexpected_input_context__unexpected_eof() -> None:
@@ -666,14 +765,16 @@ def test_lark_unexpected_input_context__unexpected_eof() -> None:
         "C": "Some c's",
     }
 
+    context = LarkUnexpectedInputContext(text, error, terminal_descriptions)
     assert """
 Expected one of: Some a's
                  Some b's
                  Some c's
 Found unexpected end of input.
 """ == to_str(
-        LarkUnexpectedInputContext(text, error, terminal_descriptions)
+        context
     )
+    check_eq(context)
 
 
 def test_lark_unexpected_input_context__unexpected_characters() -> None:
@@ -688,13 +789,15 @@ This is line 3
         "C": "Some c's",
     }
 
+    context = LarkUnexpectedInputContext(text, error, terminal_descriptions)
     assert """
 Line: "This is line 2"
                ^
 Found unexpected character.
 """ == to_str(
-        LarkUnexpectedInputContext(text, error, terminal_descriptions)
+        context
     )
+    check_eq(context)
 
 
 def test_lark_unexpected_input_context__unexpected_token() -> None:
@@ -709,61 +812,15 @@ This is line 3
         "C": "Some c's",
     }
 
+    context = LarkUnexpectedInputContext(text, error, terminal_descriptions)
     assert """
 Line:     "This is line 1"
                         ^
 Expected: Some b's
 """ == to_str(
-        LarkUnexpectedInputContext(text, error, terminal_descriptions)
+        context
     )
-
-
-@pytest.mark.parametrize(
-    "context,expected",
-    [
-        (
-            TrailingBroadcastVarrankContext("foo bar", 1, 5, "bar"),
-            """
-Line:    "foo bar"
-              ^
-Variable bar
-Broadcasting not supported for non-leading variable-rank variables.
-""",
-        ),
-        (
-            TrailingBroadcastVarrankContext("foo bar", 0, 5, "bar"),
-            """
-Variable bar
-Broadcasting not supported for non-leading variable-rank variables.
-""",
-        ),
-        (
-            TrailingBroadcastVarrankContext("foo bar", 1, 0, "bar"),
-            """
-Line:    "foo bar"
-Variable bar
-Broadcasting not supported for non-leading variable-rank variables.
-""",
-        ),
-        (
-            TrailingBroadcastVarrankContext("foo bar", 0, 0, "bar"),
-            """
-Variable bar
-Broadcasting not supported for non-leading variable-rank variables.
-""",
-        ),
-        (
-            TrailingBroadcastVarrankContext("foo bar", 1, 5, None),
-            """
-Line: "foo bar"
-           ^
-Broadcasting not supported for non-leading variable-rank variables.
-""",
-        ),
-    ],
-)
-def test_trailing_broadcast_varrank_context(context: ErrorContext, expected: str) -> None:
-    assert expected == to_str(context)
+    check_eq(context)
 
 
 @pytest.mark.parametrize(
@@ -800,3 +857,4 @@ Argument references that evaluate to multiple values are not supported for boole
 )
 def test_multiple_elements_bool_context(context: ErrorContext, expected: str) -> None:
     assert expected == to_str(context)
+    check_eq(context)
