@@ -16,11 +16,12 @@
 # pylint: disable=no-member  # PyLint struggles with TensorFlow.
 
 from time import perf_counter
-from typing import Any, Callable
+from typing import Any, Callable, Optional
 
 import numpy as np
 import pytest
 import tensorflow as tf
+from packaging.version import Version
 
 from gpflow.base import AnyNDArray
 from gpflow.experimental.check_shapes import check_shape as cs
@@ -35,7 +36,7 @@ from gpflow.experimental.check_shapes.exceptions import ShapeMismatchError
 def test_check_shapes__numpy() -> None:
     @check_shapes(
         "a: [d1, d2]",
-        "b: [d1, d3]",
+        "b: [d1, d3] if b is not None",
         "return: [d2, d3]",
     )
     def f(a: AnyNDArray, b: AnyNDArray) -> AnyNDArray:
@@ -47,7 +48,7 @@ def test_check_shapes__numpy() -> None:
 def test_check_shapes__tensorflow() -> None:
     @check_shapes(
         "a: [d1, d2]",
-        "b: [d1, d3]",
+        "b: [d1, d3] if b is not None",
         "return: [d2, d3]",
     )
     def f(a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
@@ -59,6 +60,13 @@ def test_check_shapes__tensorflow() -> None:
 def test_check_shapes__tensorflow__keras() -> None:
     # pylint: disable=arguments-differ,abstract-method,no-value-for-parameter,unexpected-keyword-arg
 
+    @check_shapes(
+        "x: [*]",
+        "return: [*]",
+    )
+    def f(x: tf.Tensor) -> tf.Tensor:
+        return x + 3
+
     class SuperLayer(tf.keras.layers.Layer):
         def __init__(self) -> None:
             super().__init__()
@@ -68,9 +76,10 @@ def test_check_shapes__tensorflow__keras() -> None:
             "x: [batch, input_dim]",
             "y: [batch, 1]",
             "return: [batch, input_dim]",
+            tf_decorator=True,
         )
         def call(self, x: tf.Tensor, y: tf.Tensor) -> tf.Tensor:
-            return x + y + self._b
+            return f(x) + y + self._b
 
     class SubLayer(SuperLayer):
         @inherit_check_shapes
@@ -85,6 +94,7 @@ def test_check_shapes__tensorflow__keras() -> None:
         @check_shapes(
             "xy: [batch, input_dim_plus_one]",
             "return: [batch, input_dim]",
+            tf_decorator=True,
         )
         def call(self, xy: tf.Tensor) -> tf.Tensor:
             x = cs(xy[:, :-1], "[batch, input_dim]")
@@ -114,49 +124,100 @@ def test_check_shapes__tensorflow__keras() -> None:
     test_layer(SubLayer())
 
 
-_F = Callable[[tf.Tensor], tf.Tensor]
+_Err = Callable[[tf.Tensor, tf.Tensor], tf.Tensor]
 _Loss = Callable[[], tf.Tensor]
 
+_ID_WRAPPER = lambda x: x
+_TF_FUNCTION = tf.function
+_SHAPED_TF_FUNCTION_ERR = tf.function(
+    input_signature=[
+        tf.TensorSpec(shape=[None], dtype=tf.float64),
+        tf.TensorSpec(shape=[], dtype=tf.float64),
+    ]
+)
+
+_SHAPED_TF_FUNCTION_LOSS = tf.function(input_signature=[])
+_UNSHAPED_TF_FUNCTION_ERR = tf.function(
+    input_signature=[
+        tf.TensorSpec(shape=None, dtype=tf.float64),
+        tf.TensorSpec(shape=None, dtype=tf.float64),
+    ]
+)
+_UNSHAPED_TF_FUNCTION_LOSS = tf.function(input_signature=[])
+_RELAXED_TF_FUNCTION = tf.function(experimental_relax_shapes=True)
+
+_NONE_SHAPE = None
+_TARGET_SHAPE = tf.TensorShape([])
+_V_SHAPE = tf.TensorShape([50])
+_UNKNOWN_SHAPE = tf.TensorShape(None)
+
 
 @pytest.mark.parametrize(
-    "f_wrapper,loss_wrapper",
+    "err_wrapper,loss_wrapper",
     [
-        (
-            lambda x: x,
-            lambda x: x,
-        ),
-        (
-            tf.function,
-            tf.function,
-        ),
-        (
-            tf.function(input_signature=[tf.TensorSpec(shape=[None], dtype=tf.float64)]),
-            tf.function(input_signature=[]),
-        ),
-        (
-            tf.function(input_signature=[tf.TensorSpec(shape=None, dtype=tf.float64)]),
-            tf.function(input_signature=[]),
-        ),
-        (
-            tf.function(experimental_relax_shapes=True),
-            tf.function(experimental_relax_shapes=True),
-        ),
+        (_ID_WRAPPER, _ID_WRAPPER),
+        (_TF_FUNCTION, _TF_FUNCTION),
+        (_SHAPED_TF_FUNCTION_ERR, _SHAPED_TF_FUNCTION_LOSS),
+        (_UNSHAPED_TF_FUNCTION_ERR, _UNSHAPED_TF_FUNCTION_LOSS),
+        (_RELAXED_TF_FUNCTION, _RELAXED_TF_FUNCTION),
     ],
 )
+@pytest.mark.parametrize("target_shape", [_NONE_SHAPE, _TARGET_SHAPE, _UNKNOWN_SHAPE])
+@pytest.mark.parametrize("v_shape", [_NONE_SHAPE, _V_SHAPE, _UNKNOWN_SHAPE])
 def test_check_shapes__tensorflow_compilation(
-    f_wrapper: Callable[[_F], _F], loss_wrapper: Callable[[_Loss], _Loss]
+    err_wrapper: Callable[[_Err], _Err],
+    loss_wrapper: Callable[[_Loss], _Loss],
+    target_shape: Optional[tf.TensorShape],
+    v_shape: Optional[tf.TensorShape],
 ) -> None:
-    target = 0.5
+    # Yeah, this test seems to be pushing the limits of TensorFlow compilation (which is probably
+    # good), but a bunch of this is fragile.
+    tf_version = Version(tf.__version__)
 
-    @f_wrapper
-    @check_shapes(
-        "x: [n]",
-        "return: [n]",
-    )
-    def f(x: tf.Tensor) -> tf.Tensor:
-        return (x - target) ** 2
+    if (target_shape is _UNKNOWN_SHAPE) or (v_shape is _UNKNOWN_SHAPE):
+        if (err_wrapper is _TF_FUNCTION) or (err_wrapper is _RELAXED_TF_FUNCTION):
+            if Version("2.7.0") <= tf_version < Version("2.8.0"):
+                pytest.skip("TensorFlow 2.7.* segfaults when trying to compile this.")
+            if Version("2.8.0") <= tf_version < Version("2.9.0"):
+                pytest.skip("TensorFlow 2.8.* is missing a TraceType(?) when trying compile this.")
 
-    v = tf.Variable(np.linspace(0.0, 1.0))
+    # See: https://github.com/tensorflow/tensorflow/issues/56414
+    if err_wrapper is _RELAXED_TF_FUNCTION:
+        if Version("2.9.0") <= tf_version < Version("2.11.0"):
+            err_wrapper = _TF_FUNCTION
+
+    if Version(tf.__version__) < Version("2.5.0"):
+        # TensorFlow < 2.5.0 doesn't like the optional `z` argument:
+
+        class SqErr:
+            @check_shapes(
+                "x: [broadcast n...]",
+                "y: [broadcast n...]",
+                "return: [n...]",
+            )
+            def __call__(self, x: tf.Tensor, y: tf.Tensor) -> tf.Tensor:
+                return (x - y) ** 2
+
+    else:
+
+        class SqErr:  # type: ignore[no-redef]
+            @check_shapes(
+                "x: [broadcast n...]",
+                "y: [broadcast n...]",
+                "z: [broadcast n...]",
+                "return: [n...]",
+            )
+            def __call__(
+                self, x: tf.Tensor, y: tf.Tensor, z: Optional[tf.Tensor] = None
+            ) -> tf.Tensor:
+                # z only declared to test the case of `None` arguments.
+                return (x - y) ** 2
+
+    sq_err = err_wrapper(SqErr())
+
+    dtype = np.float64
+    target = tf.Variable(0.5, dtype=dtype, shape=target_shape)
+    v = tf.Variable(np.linspace(0.0, 1.0), dtype=dtype, shape=v_shape)
 
     @loss_wrapper
     @check_shapes(
@@ -164,7 +225,7 @@ def test_check_shapes__tensorflow_compilation(
     )
     def loss() -> tf.Tensor:
         # keepdims is just to add an extra dimension to make the check more interesting.
-        return tf.reduce_sum(f(v), keepdims=True)
+        return tf.reduce_sum(sq_err(v, target), keepdims=True)
 
     optimiser = tf.keras.optimizers.SGD(learning_rate=0.25)
     for _ in range(10):
@@ -247,3 +308,43 @@ def test_check_shapes__disable__speed(func_wrapper: Callable[[Any], Any]) -> Non
 
     assert t_no_checks < t_with_checks
     assert t_disabled_checks < t_with_checks
+
+
+def test_issue_1864() -> None:
+    @tf.function
+    @check_shapes(
+        "x: [*]",
+        "return: [*]",
+    )
+    def f(x: tf.Tensor) -> tf.Tensor:
+        for _ in tf.range(3):
+            x = x + 1.0
+        return x
+
+    x = tf.constant(7.0)
+    f(x)
+
+
+def test_issue_1936() -> None:
+    @tf.function
+    @check_shapes(
+        "x: [*]",
+        "return: [*]",
+    )
+    def f_if(x: tf.Tensor) -> tf.Tensor:
+        if tf.size(x) == 0:
+            return x
+        else:
+            return x + x
+
+    @tf.function
+    @check_shapes(
+        "x: [*]",
+        "return: [*]",
+    )
+    def f_tf_cond(x: tf.Tensor) -> tf.Tensor:
+        return tf.cond(tf.size(x) == 0, lambda: x, lambda: x + x)
+
+    x = tf.constant(7.0)
+    f_tf_cond(x)
+    f_if(x)
