@@ -5,10 +5,15 @@ import pytest
 
 import gpflow
 from gpflow.base import AnyNDArray, RegressionData
+from gpflow.experimental.check_shapes import check_shapes
 from gpflow.models.vgp import VGP_deprecated, VGP_with_posterior
 from gpflow.posteriors import PrecomputeCacheType
 
 
+@check_shapes(
+    "regression_data[0]: [N, D]",
+    "regression_data[1]: [N, P]",
+)
 def make_models(
     regression_data: RegressionData, likelihood: gpflow.likelihoods.Likelihood
 ) -> Tuple[VGP_deprecated, VGP_with_posterior]:
@@ -22,6 +27,11 @@ def make_models(
     return mold, mnew
 
 
+@check_shapes(
+    "return[0]: [N, D]  # X",
+    "return[1]: [batch_new..., N_new, D]  # X_new",
+    "return[2]: [N, P]  # Y",
+)
 def _get_data_for_tests() -> Tuple[AnyNDArray, AnyNDArray, AnyNDArray]:
     """Helper function to create testing data"""
     X = np.random.randn(5, 6)
@@ -34,7 +44,7 @@ def _get_data_for_tests() -> Tuple[AnyNDArray, AnyNDArray, AnyNDArray]:
     "likelihood", [gpflow.likelihoods.Gaussian(), gpflow.likelihoods.Exponential()]
 )
 @pytest.mark.parametrize("full_cov", [True, False])
-@pytest.mark.parametrize("full_output_cov", [True, False])
+@pytest.mark.parametrize("full_output_cov", [False])
 def test_old_vs_new_gp_fused(
     likelihood: gpflow.likelihoods.Likelihood,
     full_cov: bool,
@@ -57,7 +67,7 @@ def test_old_vs_new_gp_fused(
     "likelihood", [gpflow.likelihoods.Gaussian(), gpflow.likelihoods.Exponential()]
 )
 @pytest.mark.parametrize("full_cov", [True, False])
-@pytest.mark.parametrize("full_output_cov", [True, False])
+@pytest.mark.parametrize("full_output_cov", [False])
 def test_old_vs_new_with_posterior(
     cache_type: PrecomputeCacheType,
     likelihood: gpflow.likelihoods.Likelihood,
